/*
 * Copyright (C) 2018-2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
/**
 *  @file
 *  contact delete
 */

#ifndef DELETE_CONTACT_HH_52AE8F85389D4E66B0EA3A26E4371466
#define DELETE_CONTACT_HH_52AE8F85389D4E66B0EA3A26E4371466

#include "libfred/opexception.hh"
#include "libfred/opcontext.hh"
#include "util/printable.hh"

#include <string>

namespace LibFred {

/**
 * Delete of contact.
 * Contact handle to delete is set via constructor.
 * Delete is executed by @ref exec method with database connection supplied in @ref OperationContext parameter.
 * In case of wrong input data or other predictable and superable failure, an instance of @ref DeleteContactByHandle::Exception is thrown with appropriate attributes set.
 * In case of other insuperable failures and inconsistencies, an instance of @ref InternalError or other exception is thrown.
 */
class DeleteContactByHandle : public Util::Printable<DeleteContactByHandle>
{
public:
    DECLARE_EXCEPTION_DATA(unknown_contact_handle, std::string);/**< exception members for unknown contact handle generated by macro @ref DECLARE_EXCEPTION_DATA*/
    DECLARE_EXCEPTION_DATA(object_linked_to_contact_handle, std::string);/**< exception members for contact linked to other object generated by macro @ref DECLARE_EXCEPTION_DATA*/
    struct Exception
    : virtual LibFred::OperationException
    , ExceptionData_unknown_contact_handle<Exception>
    , ExceptionData_object_linked_to_contact_handle<Exception>
    {};

    /**
     * Delete contact constructor with mandatory parameter.
     * @param handle sets contact identifier into @ref handle_ attribute
     */
    DeleteContactByHandle(const std::string& handle);

    /**
     * Executes delete.
     * @param ctx contains reference to database and logging interface
     */
    void exec(OperationContext& ctx);

    /**
     * Dumps state of the instance into the string
     * @return string with description of the instance state
     */
    std::string to_string()const;
private:
    const std::string handle_;/**< contact identifier */
};

/**
 * Delete of contact.
 * Id of contact to delete is set via constructor.
 * Delete is executed by @ref exec method with database connection supplied in @ref OperationContext parameter.
 * In case of wrong input data or other predictable and superable failure, an instance of @ref DeleteContactById::Exception is thrown with appropriate attributes set.
 * In case of other insuperable failures and inconsistencies, an instance of @ref InternalError or other exception is thrown.
 */
class DeleteContactById : public Util::Printable<DeleteContactById>
{
public:
    DECLARE_EXCEPTION_DATA(unknown_contact_id, unsigned long long);/**< exception members for unknown contact id generated by macro @ref DECLARE_EXCEPTION_DATA*/
    DECLARE_EXCEPTION_DATA(object_linked_to_contact_id, unsigned long long);/**< exception members for contact linked to other object generated by macro @ref DECLARE_EXCEPTION_DATA*/
    struct Exception
    : virtual LibFred::OperationException
    , ExceptionData_unknown_contact_id<Exception>
    , ExceptionData_object_linked_to_contact_id<Exception>
    {};

    /**
     * Delete contact constructor with mandatory parameter.
     * @param id sets contact identifier into @ref id_ attribute
     */
    DeleteContactById(unsigned long long id);

    /**
     * Executes delete.
     * @param ctx contains reference to database and logging interface
     */
    void exec(OperationContext& ctx);

    /**
     * Dumps state of the instance into the string
     * @return string with description of the instance state
     */
    std::string to_string()const;
private:
    const unsigned long long id_;/**< contact identifier */
};

}//namespace LibFred

#endif//DELETE_CONTACT_HH_52AE8F85389D4E66B0EA3A26E4371466
