/*
 * Copyright (C) 2018-2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */
#include "src/contact/contact_info.hh"

#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"
#include "src/fred_unwrap.hh"

#include "libfred/registrable_object/contact/contact_state.hh"
#include "libfred/registrable_object/contact/info_contact.hh"
#include "libfred/opcontext.hh"
#include "libfred/db_settings.hh"

namespace Fred {
namespace Registry {
namespace Contact {

std::string ContactInfoRequest::to_string()const
{
    return Util::StructToString().add("contact_id", contact_id)
                                 .add("contact_history_id", contact_history_id)
                                 .finish();
}

std::string ContactInfoReply::Data::to_string()const
{
    return Util::StructToString().add("contact_handle", contact_handle)
                                 .add("contact_id", contact_id)
                                 .add("contact_history_id", contact_history_id)
                                 .add("name", name)
                                 .add("organization", organization)
                                 .add("place", place)
                                 .add("telephone", telephone)
                                 .add("fax", fax)
                                 .add("emails", emails)
                                 .add("notify_emails", notify_emails)
                                 .add("vat_identification_number", vat_identification_number)
                                 .add("additional_identifier", additional_identifier)
                                 .add("mailing_address", mailing_address)
                                 .add("billing_address", billing_address)
                                 .add("shipping_address1", shipping_address[0])
                                 .add("shipping_address2", shipping_address[1])
                                 .add("shipping_address3", shipping_address[2])
                                 .add("warning_letter", warning_letter)
                                 .add("auth_info", auth_info)
                                 .add("representative_events", representative_events)
                                 .finish();
}

ContactInfoReply::Exception::ContactDoesNotExist::ContactDoesNotExist(const ContactInfoRequest& src)
    : Registry::Exception("contact " + src.to_string() + " does not exist")
{
}

ContactInfoReply::Data contact_info(const ContactInfoRequest& request)
{
    LibFred::OperationContextCreator ctx_provider;
    LibFred::OperationContextLockingForShare ctx(ctx_provider);
    try
    {
        if (request.contact_history_id != boost::none)
        {
            const auto contact_history_uuid = LibFred::RegistrableObject::make_history_uuid_of<LibFred::Object_Type::contact>(
                    Util::get_raw_value_from(*request.contact_history_id));
            return fred_unwrap(LibFred::InfoContactByHistoryUuid(contact_history_uuid).exec(ctx));
        }
        const auto contact_uuid = LibFred::RegistrableObject::make_uuid_of<LibFred::Object_Type::contact>(
                Util::get_raw_value_from(request.contact_id));
        return fred_unwrap(LibFred::InfoContactByUuid(contact_uuid).exec(ctx));
    }
    catch (const LibFred::InfoContactByUuid::Exception& e)
    {
        if (e.is_set_unknown_contact_uuid())
        {
            throw ContactInfoReply::Exception::ContactDoesNotExist(request);
        }
        throw InternalServerError("unexpected problem signalized by LibFred::InfoContactByUuid::Exception");;
    }
    catch (const LibFred::InfoContactByHistoryUuid::Exception& e)
    {
        if (e.is_set_unknown_contact_history_uuid())
        {
            throw ContactInfoReply::Exception::ContactDoesNotExist(request);
        }
        throw InternalServerError("unexpected problem signalized by LibFred::InfoContactByHistoryUuid::Exception");;
    }
    catch (const std::exception& e)
    {
        throw InternalServerError(std::string("expected unknown exception: ") + e.what());
    }
    catch (...)
    {
        throw InternalServerError("unexpected unknown exception");
    }
}

}//namespace Fred::Registry::Contact
}//namespace Fred::Registry
}//namespace Fred
