/*
 * Copyright (C) 2018-2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/contact/search_contact.hh"

#include "src/util/enum_sequence.hh"
#include "src/util/measure_duration.hh"
#include "src/util/sql_exec_in_thread.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "util/log/logger.hh"

#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <cmath>

#include <algorithm>
#include <iterator>
#include <memory>
#include <regex>
#include <set>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

namespace Fred {
namespace Registry {
namespace Contact {

std::string SearchContactRequest::to_string()const
{
    return Util::StructToString().add("query_values", query_values)
                                 .add("limit", limit)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchContactReply::Data::to_string()const
{
    return Util::StructToString().add("most_similar_contacts", most_similar_contacts)
                                 .add("result_count", result_count)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchContactReply::Data::Result::to_string()const
{
    return Util::StructToString().add("contact_id", contact_id)
                                 .add("matched_items", matched_items)
                                 .finish();
}

std::string SearchContactReply::Data::FuzzyValue::to_string()const
{
    return Util::StructToString().add("lower_estimate", lower_estimate)
                                 .add("upper_estimate", upper_estimate)
                                 .finish();
}

namespace {

std::vector<std::string> split(const std::string& src, const std::regex& delimiter)
{
    constexpr auto stuff_between_matches = -1;
    std::sregex_token_iterator begin(src.cbegin(), src.cend(), delimiter, stuff_between_matches);
    static const std::sregex_token_iterator end;
    std::vector<std::string> result;
    std::copy_if(begin, end, std::back_inserter(result), [](const std::string& item) { return !item.empty(); });
    return result;
}

std::vector<std::string> to_words(const std::string& by_spaces_delimited_words)
{
    return split(by_spaces_delimited_words, std::regex("\\s+"));
}

struct Delimiter
{
    template <typename T>
    explicit Delimiter(T&& src) : value(std::forward<T>(src)) { }
    std::string value;
};

class Collector
{
public:
    explicit Collector(Delimiter&& delimiter)
        : delimiter_(std::move(delimiter.value))
    { }
    template <typename T>
    Collector& append(T&& item)
    {
        if (!sum_.empty())
        {
            sum_.append(delimiter_);
        }
        sum_.append(std::forward<T>(item));
        return *this;
    }
    const std::string& get_sum()const noexcept
    {
        return sum_;
    }
private:
    const std::string delimiter_;
    std::string sum_;
};

class CollectorWithParentheses
{
public:
    explicit CollectorWithParentheses(Delimiter&& delimiter)
        : delimiter_(std::move(delimiter.value)),
          delimiter_in_use_(false)
    { }
    template <typename T>
    CollectorWithParentheses& append(T&& item)
    {
        if (!sum_.empty())
        {
            if (!delimiter_in_use_)
            {
                sum_ = "(" + sum_ + ")";
                delimiter_in_use_ = true;
            }
            sum_.append(delimiter_ + "(" + std::forward<T>(item) + ")");
        }
        else
        {
            sum_ = std::forward<T>(item);
        }
        return *this;
    }
    const std::string& get_sum()const noexcept
    {
        return sum_;
    }
private:
    const std::string delimiter_;
    bool delimiter_in_use_;
    std::string sum_;
};

template <ContactItem value>
struct ContactItemTraits;

template <>
struct ContactItemTraits<ContactItem::name>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(name)";
    }
    static constexpr const char* item_name()
    {
        return "name";
    }
};

template <>
struct ContactItemTraits<ContactItem::organization>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(organization)";
    }
    static constexpr const char* item_name()
    {
        return "organization";
    }
};

template <>
struct ContactItemTraits<ContactItem::place_street>
{
    static constexpr const char* transformed_column_value()
    {
        return "unaccent_streets(street1,street2,street3)";
    }
    static constexpr const char* item_name()
    {
        return "place.street";
    }
};

template <>
struct ContactItemTraits<ContactItem::place_city>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(city)";
    }
    static constexpr const char* item_name()
    {
        return "place.city";
    }
};

template <>
struct ContactItemTraits<ContactItem::place_state_or_province>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(stateorprovince)";
    }
    static constexpr const char* item_name()
    {
        return "place.state_or_province";
    }
};

template <>
struct ContactItemTraits<ContactItem::place_postal_code>
{
    static constexpr const char* transformed_column_value()
    {
        return "postalcode";
    }
    static constexpr const char* item_name()
    {
        return "place.postal_code";
    }
};

template <>
struct ContactItemTraits<ContactItem::telephone>
{
    static constexpr const char* transformed_column_value()
    {
        return "telephone";
    }
    static constexpr const char* item_name()
    {
        return "telephone";
    }
};

template <>
struct ContactItemTraits<ContactItem::fax>
{
    static constexpr const char* transformed_column_value()
    {
        return "fax";
    }
    static constexpr const char* item_name()
    {
        return "fax";
    }
};

template <>
struct ContactItemTraits<ContactItem::email>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(email)";
    }
    static constexpr const char* item_name()
    {
        return "email";
    }
};

template <>
struct ContactItemTraits<ContactItem::notify_email>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(notifyemail)";
    }
    static constexpr const char* item_name()
    {
        return "notify_email";
    }
};

template <>
struct ContactItemTraits<ContactItem::vat_identification_number>
{
    static constexpr const char* transformed_column_value()
    {
        return "vat";
    }
    static constexpr const char* item_name()
    {
        return "vat_identification_number";
    }
};

struct AggregatedSimilarity
{
    double by_words;
    double by_value;
    friend bool operator<(const AggregatedSimilarity& lhs, const AggregatedSimilarity& rhs)noexcept
    {
        return (lhs.by_words < rhs.by_words) ||
               ((lhs.by_words == rhs.by_words) && (lhs.by_value < rhs.by_value));
    }
    friend bool operator==(const AggregatedSimilarity& lhs, const AggregatedSimilarity& rhs)noexcept
    {
        return !(lhs < rhs) && !(rhs < lhs);
    }
};

struct MatchedContact
{
    unsigned long long id;
    ContactId contact_id;
    std::size_t value_length;
    std::vector<AggregatedSimilarity> similarity;
    std::set<std::string> matched_items;
};

AggregatedSimilarity get_aggregated_similarity(const Database::Result::Row& row, int column_idx, std::size_t number_of_values)
{
    static const AggregatedSimilarity absolutely_different = { 0, 0 };
    static constexpr auto offset_by_words = 0;
    static constexpr auto offset_by_value = 1;
    if (number_of_values == 1)
    {
        AggregatedSimilarity similarity;
        similarity.by_words = static_cast<double>(row[column_idx + offset_by_words]);
        similarity.by_value = static_cast<double>(row[column_idx + offset_by_value]);
        return similarity;
    }
    AggregatedSimilarity similarity = { 0, 0 };
    for (std::size_t value_idx = 0; value_idx < number_of_values; ++value_idx)
    {
        const auto by_words = static_cast<double>(row[column_idx + (2 * value_idx) + offset_by_words]);
        const auto by_value = static_cast<double>(row[column_idx + (2 * value_idx) + offset_by_value]);
        if (by_words <= 0.0)
        {
            return absolutely_different;
        }
        similarity.by_words += by_words * by_words;
        similarity.by_value += by_value * by_value;
    }
    similarity.by_words += std::sqrt(similarity.by_words / number_of_values);
    similarity.by_value += std::sqrt(similarity.by_value / number_of_values);
    return similarity;
}

void push_back_into_contacts(
        const Database::Result::Row& row,
        const std::vector<std::size_t>& numbers_of_words,
        const std::string& item_name,
        std::vector<MatchedContact>& contacts_by_id)
{
    static constexpr auto column_id = 0;
    static constexpr auto column_uuid = 1;
    static constexpr auto column_length = 2;
    static constexpr auto column_similarity_start = 3;
    MatchedContact match;
    match.id = static_cast<unsigned long long>(row[column_id]);
    const bool check_order = !contacts_by_id.empty();
    if (check_order)
    {
        const bool order_is_correct = contacts_by_id[contacts_by_id.size() - 1].id < match.id;
        if (!order_is_correct)
        {
            struct IncorrectOrder : std::runtime_error
            {
                IncorrectOrder() : std::runtime_error("Incorrect order of matched contacts'") { }
            };
            throw IncorrectOrder();
        }
    }
    match.contact_id = Util::make_strong<Registry::Contact::ContactId>(
            boost::uuids::string_generator()(static_cast<std::string>(row[column_uuid])));
    match.value_length = static_cast<std::size_t>(row[column_length]);
    match.matched_items.insert(item_name);
    match.similarity.reserve(numbers_of_words.size());
    auto column_idx = column_similarity_start;
    for (const auto words : numbers_of_words)
    {
        match.similarity.push_back(get_aggregated_similarity(row, column_idx, words));
        column_idx += 2 * words;
    }
    contacts_by_id.push_back(match);
}

//SELECT c.id,
//       obr.uuid,
//       length(c.v),
//       word_similarity(unaccented.v1w1,c.v),similarity(unaccented.v1w1,c.v),
//       word_similarity(unaccented.v1w2,c.v),similarity(unaccented.v1w2,c.v),
//       word_similarity(unaccented.v2w1,c.v),similarity(unaccented.v2w1,c.v),
//FROM (SELECT f_unaccent('Moravské'),f_unaccent('Budějovice'),f_unaccent('Novák')) AS unaccented(v1w1,v1w2,v2w1),
//     (SELECT id,f_unaccent(city) FROM contact) AS c(id,v)
//JOIN object_registry obr ON obr.id=c.id
//WHERE ((unaccented.v1w1 <% c.v) AND
//       (unaccented.v1w2 <% c.v)) OR
//      (unaccented.v2w1 <% c.v)
//ORDER BY 1;
template <ContactItem item>
class SearchIn
{
public:
    explicit SearchIn(const std::vector<std::size_t>& numbers_of_words)
    {
        Collector what(Delimiter(","));
        Collector unaccented_what(Delimiter(","));
        Collector unaccented_as(Delimiter(","));
        CollectorWithParentheses where(Delimiter(" OR "));
        int param_idx = 0;
        int value_idx = 0;
        for (const auto words : numbers_of_words)
        {
            CollectorWithParentheses sub_where(Delimiter(" AND "));
            const std::string value_name = "v" + std::to_string(value_idx + 1) + "w";
            for (unsigned word_idx = 0; word_idx < words; ++word_idx, ++param_idx)
            {
                const std::string value_word_name = value_name + std::to_string(word_idx + 1);
                what.append("word_similarity(unaccented." + value_word_name + ",c.v),"
                            "similarity(unaccented." + value_word_name + ",c.v)");
                unaccented_what.append("f_unaccent($" + std::to_string(param_idx + 1) + "::TEXT)");
                unaccented_as.append(value_word_name);
                sub_where.append("unaccented." + value_word_name + " <% c.v");
            }
            where.append(sub_where.get_sum());
            ++value_idx;
        }
        sql_ = "SELECT c.id,"
                      "obr.uuid,"
                      "length(c.v)," +
                       what.get_sum() + " "
               "FROM (SELECT " + unaccented_what.get_sum() + ") unaccented(" + unaccented_as.get_sum() + "),"
                    "(SELECT id," + ContactItemTraits<item>::transformed_column_value() + " FROM contact) c(id,v) "
               "JOIN object_registry obr ON obr.id=c.id "
               "WHERE " + where.get_sum() + " "
               "ORDER BY 1";
    }
    const std::string& get_sql()const noexcept
    {
        return sql_;
    }
    static void push_back(const std::vector<std::size_t>& numbers_of_words,
                          const Database::Result::Row& row,
                          std::vector<MatchedContact>& contacts_by_id)
    {
        push_back_into_contacts(row, numbers_of_words, ContactItemTraits<item>::item_name(), contacts_by_id);
    }
private:
    std::string sql_;
};

template <>
class SearchIn<ContactItem::contact_handle>
{
public:
    explicit SearchIn(const std::vector<std::size_t>& numbers_of_words)
    {
        Collector what(Delimiter(","));
        Collector unaccented_what(Delimiter(","));
        Collector unaccented_as(Delimiter(","));
        CollectorWithParentheses where(Delimiter(" OR "));
        int param_idx = 0;
        int value_idx = 0;
        for (const auto words : numbers_of_words)
        {
            CollectorWithParentheses sub_where(Delimiter(" AND "));
            const std::string value_name = "v" + std::to_string(value_idx + 1) + "w";
            for (unsigned word_idx = 0; word_idx < words; ++word_idx, ++param_idx)
            {
                const std::string value_word_name = value_name + std::to_string(word_idx + 1);
                what.append("word_similarity(unaccented." + value_word_name + ",obr.name),"
                            "similarity(unaccented." + value_word_name + ",obr.name)");
                unaccented_what.append("f_unaccent($" + std::to_string(param_idx + 1) + "::TEXT)");
                unaccented_as.append(value_word_name);
                sub_where.append("unaccented." + value_word_name + " <% obr.name");
            }
            where.append(sub_where.get_sum());
            ++value_idx;
        }
        sql_ = "SELECT c.id,"
                      "obr.uuid,"
                      "length(obr.name)," +
                       what.get_sum() + " "
               "FROM (SELECT " + unaccented_what.get_sum() + ") unaccented(" + unaccented_as.get_sum() + ")," +
                    "object_registry obr "
               "JOIN contact c ON c.id=obr.id "
               "WHERE (" + where.get_sum() + ") AND "
                     "obr.erdate IS NULL "
               "ORDER BY 1";
    }
    const std::string& get_sql()const noexcept
    {
        return sql_;
    }
    static void push_back(const std::vector<std::size_t>& numbers_of_words,
                          const Database::Result::Row& row,
                          std::vector<MatchedContact>& contacts_by_id)
    {
        push_back_into_contacts(row, numbers_of_words, "contact_handle", contacts_by_id);
    }
private:
    std::string sql_;
};

template <>
class SearchIn<ContactItem::additional_identifier>
{
public:
    explicit SearchIn(const std::vector<std::size_t>& numbers_of_words)
    {
        Collector what(Delimiter(","));
        Collector unaccented_what(Delimiter(","));
        Collector unaccented_as(Delimiter(","));
        CollectorWithParentheses where(Delimiter(" OR "));
        int param_idx = 0;
        int value_idx = 0;
        for (const auto words : numbers_of_words)
        {
            CollectorWithParentheses sub_where(Delimiter(" AND "));
            const std::string value_name = "v" + std::to_string(value_idx + 1) + "w";
            for (unsigned word_idx = 0; word_idx < words; ++word_idx, ++param_idx)
            {
                const std::string value_word_name = value_name + std::to_string(word_idx + 1);
                what.append("word_similarity(unaccented." + value_word_name + ",c.v),"
                            "similarity(unaccented." + value_word_name + ",c.v)");
                unaccented_what.append("f_unaccent($" + std::to_string(param_idx + 1) + "::TEXT)");
                unaccented_as.append(value_word_name);
                sub_where.append("unaccented." + value_word_name + " <% c.v");
            }
            where.append(sub_where.get_sum());
            ++value_idx;
        }
        sql_ = "SELECT c.id,"
                      "c.ssn_type,"
                      "obr.uuid,"
                      "length(c.v)," +
                       what.get_sum() + " "
               "FROM (SELECT " + unaccented_what.get_sum() + ") unaccented(" + unaccented_as.get_sum() + "),"
                    "(SELECT c.id,c.ssn,es.type "
                     "FROM contact c "
                     "JOIN enum_ssntype es ON es.id=c.ssntype) c(id,v,ssn_type) "
               "JOIN object_registry obr ON obr.id=c.id "
               "WHERE " + where.get_sum() + " "
               "ORDER BY 1";
    }
    const std::string& get_sql()const noexcept
    {
        return sql_;
    }
    static void push_back(const std::vector<std::size_t>& numbers_of_words,
                          const Database::Result::Row& row,
                          std::vector<MatchedContact>& contacts_by_id)
    {
        static constexpr auto column_id = 0;
        static constexpr auto column_ssn_type = 1;
        static constexpr auto column_uuid = 2;
        static constexpr auto column_length = 3;
        static constexpr auto column_similarity_start = 4;
        MatchedContact match;
        match.id = static_cast<unsigned long long>(row[column_id]);
        const bool check_order = !contacts_by_id.empty();
        if (check_order)
        {
            const bool order_is_correct = contacts_by_id[contacts_by_id.size() - 1].id < match.id;
            if (!order_is_correct)
            {
                struct IncorrectOrder : std::runtime_error
                {
                    IncorrectOrder() : std::runtime_error("Incorrect order of matched contacts'") { }
                };
                throw IncorrectOrder();
            }
        }
        const auto ssn_type = row[column_ssn_type].isnull() ? std::string()
                                                            : static_cast<std::string>(row[column_ssn_type]);
        std::string ssn_type_name;
        if (ssn_type == "RC")
        {
            ssn_type_name = "national_identity_number";
        }
        else if (ssn_type == "OP")
        {
            ssn_type_name = "national_identity_card";
        }
        else if (ssn_type == "PASS")
        {
            ssn_type_name = "passport_number";
        }
        else if (ssn_type == "ICO")
        {
            ssn_type_name = "company_registration_number";
        }
        else if (ssn_type == "MPSV")
        {
            ssn_type_name = "social_security_number";
        }
        else if (ssn_type == "BIRTHDAY")
        {
            ssn_type_name = "birthdate";
        }
        match.contact_id = Util::make_strong<Registry::Contact::ContactId>(
                boost::uuids::string_generator()(static_cast<std::string>(row[column_uuid])));
        match.value_length = static_cast<std::size_t>(row[column_length]);
        match.matched_items.insert("additional_identifier." + ssn_type_name);
        match.similarity.reserve(numbers_of_words.size());
        auto column_idx = column_similarity_start;
        for (const auto words : numbers_of_words)
        {
            match.similarity.push_back(get_aggregated_similarity(row, column_idx, words));
            column_idx += 2 * words;
        }
        contacts_by_id.push_back(match);
    }
private:
    std::string sql_;
};

template <>
struct ContactItemTraits<ContactItem::address_company>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(company_name)";
    }
    static constexpr const char* item_name()
    {
        return "company";
    }
};

template <>
struct ContactItemTraits<ContactItem::address_street>
{
    static constexpr const char* transformed_column_value()
    {
        return "unaccent_streets(street1,street2,street3)";
    }
    static constexpr const char* item_name()
    {
        return "street";
    }
};

template <>
struct ContactItemTraits<ContactItem::address_city>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(city)";
    }
    static constexpr const char* item_name()
    {
        return "city";
    }
};

template <>
struct ContactItemTraits<ContactItem::address_state_or_province>
{
    static constexpr const char* transformed_column_value()
    {
        return "f_unaccent(stateorprovince)";
    }
    static constexpr const char* item_name()
    {
        return "state_or_province";
    }
};

template <>
struct ContactItemTraits<ContactItem::address_postal_code>
{
    static constexpr const char* transformed_column_value()
    {
        return "postalcode";
    }
    static constexpr const char* item_name()
    {
        return "postal_code";
    }
};

template <ContactItem item>
class SearchInContactAddress
{
public:
    explicit SearchInContactAddress(const std::vector<std::size_t>& numbers_of_words)
    {
        Collector what(Delimiter(","));
        Collector unaccented_what(Delimiter(","));
        Collector unaccented_as(Delimiter(","));
        CollectorWithParentheses where(Delimiter(" OR "));
        int param_idx = 0;
        int value_idx = 0;
        for (const auto words : numbers_of_words)
        {
            CollectorWithParentheses sub_where(Delimiter(" AND "));
            const std::string value_name = "v" + std::to_string(value_idx + 1) + "w";
            for (unsigned word_idx = 0; word_idx < words; ++word_idx, ++param_idx)
            {
                const std::string value_word_name = value_name + std::to_string(word_idx + 1);
                what.append("word_similarity(unaccented." + value_word_name + ",c.v),"
                            "similarity(unaccented." + value_word_name + ",c.v)");
                unaccented_what.append("f_unaccent($" + std::to_string(param_idx + 1) + "::TEXT)");
                unaccented_as.append(value_word_name);
                sub_where.append("unaccented." + value_word_name + " <% c.v");
            }
            where.append(sub_where.get_sum());
            ++value_idx;
        }
        sql_ = "SELECT c.id,"
                      "c.type,"
                      "obr.uuid,"
                      "length(c.v)," +
                       what.get_sum() + " "
               "FROM (SELECT " + unaccented_what.get_sum() + ") unaccented(" + unaccented_as.get_sum() + "),"
                    "(SELECT contactid," + ContactItemTraits<item>::transformed_column_value() + ",type "
                     "FROM contact_address) c(id,v,type) "
               "JOIN object_registry obr ON obr.id=c.id "
               "WHERE " + where.get_sum() + " "
               "ORDER BY 1";
    }
    const std::string& get_sql()const noexcept
    {
        return sql_;
    }
    static void push_back(const std::vector<std::size_t>& numbers_of_words,
                          const Database::Result::Row& row,
                          std::vector<MatchedContact>& contacts_by_id)
    {
        static constexpr auto column_id = 0;
        static constexpr auto column_address_type = 1;
        static constexpr auto column_uuid = 2;
        static constexpr auto column_length = 3;
        static constexpr auto column_similarity_start = 4;
        MatchedContact match;
        match.id = static_cast<unsigned long long>(row[column_id]);
        const bool check_order = !contacts_by_id.empty();
        if (check_order)
        {
            const bool order_is_correct = contacts_by_id[contacts_by_id.size() - 1].id <= match.id;
            if (!order_is_correct)
            {
                struct IncorrectOrder : std::runtime_error
                {
                    IncorrectOrder() : std::runtime_error("Incorrect order of matched contacts'") { }
                };
                throw IncorrectOrder();
            }
        }
        const auto address_type = static_cast<std::string>(row[column_address_type]);
        std::string item_name = ContactItemTraits<item>::item_name();
        if (address_type == "MAILING")
        {
            item_name = "mailing_address." + item_name;
        }
        else if (address_type == "BILLING")
        {
            item_name = "billing_address." + item_name;
        }
        else if (address_type == "SHIPPING")
        {
            item_name = "shipping_address1." + item_name;
        }
        else if (address_type == "SHIPPING_2")
        {
            item_name = "shipping_address2." + item_name;
        }
        else if (address_type == "SHIPPING_3")
        {
            item_name = "shipping_address3." + item_name;
        }
        else
        {
            item_name = "." + item_name;
        }
        match.contact_id = Util::make_strong<Registry::Contact::ContactId>(
                boost::uuids::string_generator()(static_cast<std::string>(row[column_uuid])));
        match.value_length = static_cast<std::size_t>(row[column_length]);
        match.matched_items.insert(item_name);
        match.similarity.reserve(numbers_of_words.size());
        auto column_idx = column_similarity_start;
        for (const auto words : numbers_of_words)
        {
            match.similarity.push_back(get_aggregated_similarity(row, column_idx, words));
            column_idx += 2 * words;
        }
        if (contacts_by_id[contacts_by_id.size() - 1].id == match.id)
        {
            auto& last_contact = contacts_by_id[contacts_by_id.size() - 1];
            last_contact.matched_items.insert(item_name);
            for (std::size_t idx = 0; idx < match.similarity.size(); ++idx)
            {
                if (last_contact.similarity[idx] < match.similarity[idx])
                {
                    last_contact.similarity[idx] = match.similarity[idx];
                }
            }
            return;
        }
        contacts_by_id.push_back(match);
    }
private:
    std::string sql_;
};

template <>
class SearchIn<ContactItem::address_company> : public SearchInContactAddress<ContactItem::address_company>
{
public:
    using SearchInContactAddress<ContactItem::address_company>::SearchInContactAddress;
};

template <>
class SearchIn<ContactItem::address_street> : public SearchInContactAddress<ContactItem::address_street>
{
public:
    using SearchInContactAddress<ContactItem::address_street>::SearchInContactAddress;
};

template <>
class SearchIn<ContactItem::address_city> : public SearchInContactAddress<ContactItem::address_city>
{
public:
    using SearchInContactAddress<ContactItem::address_city>::SearchInContactAddress;
};

template <>
class SearchIn<ContactItem::address_state_or_province> : public SearchInContactAddress<ContactItem::address_state_or_province>
{
public:
    using SearchInContactAddress<ContactItem::address_state_or_province>::SearchInContactAddress;
};

template <>
class SearchIn<ContactItem::address_postal_code> : public SearchInContactAddress<ContactItem::address_postal_code>
{
public:
    using SearchInContactAddress<ContactItem::address_postal_code>::SearchInContactAddress;
};

std::unique_ptr<std::vector<MatchedContact>> combine(const std::vector<MatchedContact>& a, const std::vector<MatchedContact>& b)
{
    auto sum = std::make_unique<std::vector<MatchedContact>>();
    sum->reserve(a.size() + b.size());
    auto a_top = a.begin();
    auto b_top = b.begin();
    while ((a_top != a.end()) && (b_top != b.end()))
    {
        if (a_top->id < b_top->id)
        {
            sum->push_back(*a_top);
            ++a_top;
        }
        else if (b_top->id < a_top->id)
        {
            sum->push_back(*b_top);
            ++b_top;
        }
        else
        {
            auto c = *a_top;
            c.matched_items.insert(b_top->matched_items.begin(), b_top->matched_items.end());
            for (std::size_t idx = 0; idx < c.similarity.size(); ++idx)
            {
                if (c.similarity[idx] < b_top->similarity[idx])
                {
                    c.similarity[idx] = b_top->similarity[idx];
                }
            }
            sum->push_back(c);
            ++a_top;
            ++b_top;
        }
    }
    while (a_top != a.end())
    {
        sum->push_back(*a_top);
        ++a_top;
    }
    while (b_top != b.end())
    {
        sum->push_back(*b_top);
        ++b_top;
    }
    return sum;
}

template <ContactItem ...items>
using ContactColumns = std::integer_sequence<ContactItem, items...>;

template <typename T>
using SearchableItems = Util::EnumSequence<ContactColumns<ContactItem::name,
                                                          ContactItem::organization,
                                                          ContactItem::place_street,
                                                          ContactItem::place_city,
                                                          ContactItem::place_state_or_province,
                                                          ContactItem::place_postal_code,
                                                          ContactItem::telephone,
                                                          ContactItem::fax,
                                                          ContactItem::email,
                                                          ContactItem::notify_email,
                                                          ContactItem::vat_identification_number,
                                                          ContactItem::additional_identifier,
                                                          ContactItem::contact_handle,
                                                          ContactItem::address_company,
                                                          ContactItem::address_street,
                                                          ContactItem::address_city,
                                                          ContactItem::address_state_or_province,
                                                          ContactItem::address_postal_code>,
                                           T>;

class RunQuery
{
public:
    RunQuery(const std::set<ContactItem>& searched_items,
             const std::vector<std::size_t>& numbers_of_words,
             const Database::QueryParams& params)
        : searched_items_{searched_items},
          numbers_of_words_{numbers_of_words},
          params_{params}
    { }
    template <ContactItem column>
    void operator()(Util::SqlInThreadSolver& running_sql)const
    {
        const bool to_search = 0 < searched_items_.count(column);
        if (to_search)
        {
            running_sql = Util::sql_exec_in_thread(SearchIn<column>{numbers_of_words_}.get_sql(), params_);
        }
    }
private:
    const std::set<ContactItem>& searched_items_;
    const std::vector<std::size_t>& numbers_of_words_;
    const Database::QueryParams& params_;
};

class GetResult
{
public:
    explicit GetResult(const std::vector<std::size_t>& numbers_of_words)
        : numbers_of_words_{numbers_of_words},
        sum_matches_{std::make_unique<std::vector<MatchedContact>>()}
    { }
    template <ContactItem column>
    void operator()(Util::SqlInThreadSolver& running_sql)
    {
        if (running_sql.is_joinable())
        {
            const Database::Result task_result = Util::SqlInThreadSolver::Result(std::move(running_sql)).get_result();
            if (0 < task_result.size())
            {
                auto matches = std::make_unique<std::vector<MatchedContact>>();
                matches->reserve(task_result.size());
                for (std::size_t row_idx = 0; row_idx < task_result.size(); ++row_idx)
                {
                    SearchIn<column>::push_back(numbers_of_words_, task_result[row_idx], *matches);
                }
                if (sum_matches_->empty())
                {
                    sum_matches_ = std::move(matches);
                }
                else
                {
                    auto combined_matches = combine(*sum_matches_, *matches);
                    sum_matches_ = std::move(combined_matches);
                }
            }
        }
    }
    const std::vector<MatchedContact>& get_sum_matches()const noexcept
    {
        return *sum_matches_;
    }
    void release_sum_matches()noexcept
    {
        sum_matches_.release();
    }
private:
    const std::vector<std::size_t>& numbers_of_words_;
    std::unique_ptr<std::vector<MatchedContact>> sum_matches_;
};

}//namespace Fred::Registry::Contact::{anonymous}

SearchContactReply::Data search_contact(const SearchContactRequest& request)
{
    Database::QueryParams params;
    const std::vector<std::size_t> numbers_of_words = [&]()
    {
        std::vector<std::size_t> numbers_of_words;
        numbers_of_words.reserve(request.query_values.size());
        for (const auto& by_spaces_delimited_words : request.query_values)
        {
            const auto words = to_words(by_spaces_delimited_words);
            if (!words.empty())
            {
                params.insert(params.end(), words.cbegin(), words.cend());
            }
            numbers_of_words.push_back(words.size());
        }
        return numbers_of_words;
    }();
    SearchableItems<Util::SqlInThreadSolver> running_sql{};
    visit(running_sql, RunQuery(request.searched_items, numbers_of_words, params));
    GetResult data_collector{numbers_of_words};
    {
        const struct LogDuration : Util::MeasureDuration<LogDuration>
        {
            static void measure_completed(Time duration_ms)
            {
                LOGGER.debug("data collected in " + std::to_string(duration_ms) + "ms");
            }
        } log_duration{};
        visit(running_sql, data_collector);
    }
    const struct LogDuration : Util::MeasureDuration<LogDuration>
    {
        static void measure_completed(Time duration_ms)
        {
            LOGGER.debug("matches completed in " + std::to_string(duration_ms) + "ms");
        }
    } log_duration{};
    struct ResultWithSimilarity
    {
        explicit ResultWithSimilarity(const MatchedContact& src)
            : similarity(compute_similarity(src)),
              value_length(src.value_length)
        {
            if (0.0 < similarity.by_words)
            {
                result.contact_id = src.contact_id;
                result.matched_items = src.matched_items;
            }
        }
        SearchContactReply::Data::Result result;
        AggregatedSimilarity similarity;
        std::size_t value_length;
        static AggregatedSimilarity compute_similarity(const MatchedContact& src)
        {
            if (src.similarity.size() == 1)
            {
                return src.similarity[0];
            }
            static constexpr AggregatedSimilarity absolutely_different = { 0, 0 };
            AggregatedSimilarity sum = { 0, 0 };
            for (const auto similarity : src.similarity)
            {
                if (similarity.by_words <= 0.0)
                {
                    return absolutely_different;
                }
                sum.by_words += similarity.by_words * similarity.by_words;
                sum.by_value += similarity.by_value * similarity.by_value;
            }
            sum.by_words = std::sqrt(sum.by_words / src.similarity.size());
            sum.by_value = std::sqrt(sum.by_value / src.similarity.size());
            return sum;
        }
    };
    std::vector<ResultWithSimilarity> matches_by_similarity{};
    matches_by_similarity.reserve(data_collector.get_sum_matches().size());
    for (const auto& matched_contact : data_collector.get_sum_matches())
    {
        const ResultWithSimilarity item(matched_contact);
        if (0.0 < item.similarity.by_words)
        {
            matches_by_similarity.push_back(item);
        }
    }
    data_collector.release_sum_matches();
    struct BySimilarityDescByLengthAsc
    {
        bool operator()(const ResultWithSimilarity& lhs, const ResultWithSimilarity& rhs)const
        {
            return (rhs.similarity < lhs.similarity) ||
                   ((rhs.similarity == lhs.similarity) && (lhs.value_length < rhs.value_length));
        }
    };
    std::sort(matches_by_similarity.begin(), matches_by_similarity.end(), BySimilarityDescByLengthAsc());
    SearchContactReply::Data result;
    result.result_count.lower_estimate = matches_by_similarity.size();
    result.result_count.upper_estimate = matches_by_similarity.size();
    result.searched_items = request.searched_items;
    const bool unlimited = request.limit == boost::none;
    if (unlimited || (matches_by_similarity.size() <= *request.limit))
    {
        result.most_similar_contacts.reserve(matches_by_similarity.size());
        for (const auto& item : matches_by_similarity)
        {
            result.most_similar_contacts.push_back(item.result);
        }
        return result;
    }
    result.most_similar_contacts.reserve(*request.limit);
    for (const auto& item : matches_by_similarity)
    {
        if (*request.limit <= result.most_similar_contacts.size())
        {
            return result;
        }
        result.most_similar_contacts.push_back(item.result);
    }
    return result;
}

}//namespace Fred::Registry::Contact
}//namespace Fred::Registry
}//namespace Fred
