/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/domain/search_domain.hh"

#include "src/util/sql_exec_in_thread.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "util/log/logger.hh"

#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <stdexcept>
#include <string>

namespace Fred {
namespace Registry {
namespace Domain {

std::string SearchDomainRequest::to_string() const
{
    return Util::StructToString().add("query_values", query_values)
                                 .add("limit", limit)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchDomainReply::Data::to_string() const
{
    return Util::StructToString().add("most_similar_domains", most_similar_domains)
                                 .add("result_count", result_count)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchDomainReply::Data::Result::to_string() const
{
    return Util::StructToString().add("domain_id", domain_id)
                                 .add("matches", matches)
                                 .add("matched_items", matched_items)
                                 .finish();
}

std::string SearchDomainReply::Data::Result::Match::to_string() const
{
    return Util::StructToString().add("similarity", similarity)
                                 .add("item", item)
                                 .finish();
}

std::string SearchDomainReply::Data::FuzzyValue::to_string() const
{
    return Util::StructToString().add("lower_estimate", lower_estimate)
                                 .add("upper_estimate", upper_estimate)
                                 .finish();
}

namespace {

auto get_result_from_match(const Database::Result::Row& row)
{
    struct
    {
        DomainId domain_id;
        float similarity;
    } result;
    result.domain_id =
        Util::make_strong<Registry::Domain::DomainId>(
            boost::uuids::string_generator()(static_cast<std::string>(row[0])));
    result.similarity = static_cast<decltype(result.similarity)>(row[1]);
    return result;
}

}//namespace Fred::Registry::Domain::{anonymous}

SearchDomainReply::Data search_domain(const SearchDomainRequest& request)
{
    static const std::string sql =
        "WITH unaccented AS (SELECT f_unaccent($1::TEXT) AS query) "
        "SELECT uuid,similarity(name,unaccented.query) "
        "FROM unaccented,object_registry "
        "WHERE name % unaccented.query AND "
        "type=get_object_type_id('domain') AND "
        "erdate IS NULL "
        "ORDER BY 2 DESC,length(name),name,id DESC";
    if (request.query_values.size() != 1)
    {
        throw std::runtime_error("just one value must be searched");
    }
    const Database::QueryParams params = { request.query_values[0] };
    const Database::Result match{Util::sql_exec(sql, params)};
    SearchDomainReply::Data result;
    result.result_count.lower_estimate = match.size();
    result.result_count.upper_estimate = match.size();
    result.searched_items.insert(DomainItem::fqdn);
    const bool unlimited = request.limit == boost::none;
    const auto matches = (unlimited || (match.size() <= *request.limit)) ? match.size() : *request.limit;
    result.most_similar_domains.reserve(matches);
    for (std::size_t idx = 0; idx < matches; ++idx)
    {
        const auto item_match_result = get_result_from_match(match[idx]);
        SearchDomainReply::Data::Result domain;
        SearchDomainReply::Data::Result::Match col_match;
        col_match.similarity = item_match_result.similarity;
        col_match.item = DomainItem::fqdn;
        domain.domain_id = item_match_result.domain_id;
        domain.matches.push_back(col_match);
        domain.matched_items.insert(DomainItem::fqdn);
        result.most_similar_domains.push_back(domain);
    }
    return result;
}

}//namespace Fred::Registry::Domain
}//namespace Fred::Registry
}//namespace Fred
