/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "src/keyset/search_keyset.hh"

#include "src/util/enum_sequence.hh"
#include "src/util/measure_duration.hh"
#include "src/util/sql_exec_in_thread.hh"
#include "src/util/strong_type.hh"
#include "src/util/struct_to_string.hh"

#include "util/log/logger.hh"

#include <boost/uuid/string_generator.hpp>
#include <boost/uuid/uuid_io.hpp>

#include <memory>
#include <set>
#include <stdexcept>
#include <string>
#include <utility>
#include <vector>

namespace Fred {
namespace Registry {
namespace Keyset {

std::string SearchKeysetRequest::to_string() const
{
    return Util::StructToString().add("query_values", query_values)
                                 .add("limit", limit)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchKeysetReply::Data::to_string() const
{
    return Util::StructToString().add("most_similar_keysets", most_similar_keysets)
                                 .add("result_count", result_count)
                                 .add("searched_items", searched_items)
                                 .finish();
}

std::string SearchKeysetReply::Data::Result::to_string() const
{
    return Util::StructToString().add("keyset_id", keyset_id)
                                 .add("matched_items", matched_items)
                                 .finish();
}

std::string SearchKeysetReply::Data::FuzzyValue::to_string() const
{
    return Util::StructToString().add("lower_estimate", lower_estimate)
                                 .add("upper_estimate", upper_estimate)
                                 .finish();
}

namespace {

struct Delimiter
{
    template <typename T>
    explicit Delimiter(T&& src) : value(std::forward<T>(src)) { }
    std::string value;
};

class Collector
{
public:
    explicit Collector(Delimiter&& delimiter)
        : delimiter_(std::move(delimiter.value))
    { }
    template <typename T>
    Collector& append(T&& item)
    {
        if (!sum_.empty())
        {
            sum_.append(delimiter_);
        }
        sum_.append(std::forward<T>(item));
        return *this;
    }
    const std::string& get_sum()const noexcept
    {
        return sum_;
    }
private:
    const std::string delimiter_;
    std::string sum_;
};

class CollectorWithParentheses
{
public:
    explicit CollectorWithParentheses(Delimiter&& delimiter)
        : delimiter_(std::move(delimiter.value)),
          delimiter_in_use_(false)
    { }
    template <typename T>
    CollectorWithParentheses& append(T&& item)
    {
        if (!sum_.empty())
        {
            if (!delimiter_in_use_)
            {
                sum_ = "(" + sum_ + ")";
                delimiter_in_use_ = true;
            }
            sum_.append(delimiter_ + "(" + std::forward<T>(item) + ")");
        }
        else
        {
            sum_ = std::forward<T>(item);
        }
        return *this;
    }
    const std::string& get_sum()const noexcept
    {
        return sum_;
    }
private:
    const std::string delimiter_;
    bool delimiter_in_use_;
    std::string sum_;
};

struct MatchedKeyset
{
    unsigned long long id;
    KeysetId keyset_id;
    std::vector<double> similarity;
    std::size_t value_length;
    std::set<KeysetItem> matched_items;
};

template <KeysetItem item>
class SearchIn;

//SELECT k.id,
//       obr.uuid,
//       length(obr.name),
//       similarity(unaccented.v1,obr.name),
//       similarity(unaccented.v2,obr.name)
//FROM (SELECT f_unaccent('ABCD'),f_unaccent('EFGH')) AS unaccented(v1,v2),
//     object_registry obr
//JOIN keyset k ON k.id=obr.id
//WHERE ((unaccented.v1 % obr.name) OR
//       (unaccented.v2 % obr.name)) AND obr.erdate IS NULL
//ORDER BY 1;
template <>
class SearchIn<KeysetItem::keyset_handle>
{
public:
    explicit SearchIn(std::size_t number_of_values)
    {
        Collector what(Delimiter(","));
        Collector unaccented_what(Delimiter(","));
        Collector unaccented_as(Delimiter(","));
        CollectorWithParentheses where(Delimiter(" OR "));
        for (std::size_t idx = 0; idx < number_of_values; ++idx)
        {
            const std::string value_name = "v" + std::to_string(idx + 1);
            what.append("similarity(unaccented." + value_name + ",obr.name)");
            unaccented_what.append("f_unaccent($" + std::to_string(idx + 1) + "::TEXT)");
            unaccented_as.append(value_name);
            where.append("unaccented." + value_name + " % obr.name");
        }
        sql_ = "SELECT k.id,"
                      "obr.uuid,"
                      "length(obr.name)," +
                       what.get_sum() + " "
               "FROM (SELECT " + unaccented_what.get_sum() + ") unaccented(" + unaccented_as.get_sum() + ")," +
                    "object_registry obr "
               "JOIN keyset k ON k.id=obr.id "
               "WHERE (" + where.get_sum() + ") AND "
                     "obr.erdate IS NULL "
               "ORDER BY 1";
    }
    const std::string& get_sql()const noexcept
    {
        return sql_;
    }
    static void push_back(const Database::Result::Row& row, std::vector<MatchedKeyset>& keysets_by_id)
    {
        static constexpr auto column_id = 0;
        static constexpr auto column_uuid = 1;
        static constexpr auto column_length = 2;
        static constexpr auto column_similarity_start = 3;
        MatchedKeyset match;
        match.id = static_cast<unsigned long long>(row[column_id]);
        const bool check_order = !keysets_by_id.empty();
        if (check_order)
        {
            const bool order_is_correct = keysets_by_id[keysets_by_id.size() - 1].id < match.id;
            if (!order_is_correct)
            {
                struct IncorrectOrder : std::runtime_error
                {
                    IncorrectOrder() : std::runtime_error("Incorrect order of matched keysets'") { }
                };
                throw IncorrectOrder();
            }
        }
        match.keyset_id = Util::make_strong<Registry::Keyset::KeysetId>(
            boost::uuids::string_generator()(static_cast<std::string>(row[column_uuid])));
        match.matched_items.insert(KeysetItem::keyset_handle);
        match.similarity.reserve(row.size());
        for (decltype(row.size()) column_idx = column_similarity_start; column_idx < row.size(); ++column_idx)
        {
            match.similarity.push_back(static_cast<double>(row[column_idx]));
        }
        match.value_length = static_cast<std::size_t>(row[column_length]);
        keysets_by_id.push_back(match);
    }
private:
    std::string sql_;
};

//SELECT k.id,
//       obr.uuid,
//       searched.v1=d.key,
//       searched.v2=d.key
//FROM (SELECT 'foobar'::TEXT,'barfoo'::TEXT) AS searched(v1,v2),
//     keyset k
//JOIN dnskey d ON d.keysetid=k.id
//JOIN object_registry obr ON obr.id=k.id
//WHERE searched.v1=d.key OR
//      searched.v2=d.key
//ORDER BY 1;
template <>
class SearchIn<KeysetItem::public_key>
{
public:
    explicit SearchIn(std::size_t number_of_values)
    {
        Collector what(Delimiter(","));
        Collector searched_what(Delimiter(","));
        Collector searched_as(Delimiter(","));
        CollectorWithParentheses where(Delimiter(" OR "));
        for (std::size_t idx = 0; idx < number_of_values; ++idx)
        {
            const std::string value_name = "v" + std::to_string(idx + 1);
            what.append("searched." + value_name + "=d.key");
            searched_what.append("$" + std::to_string(idx + 1) + "::TEXT");
            searched_as.append(value_name);
            where.append("searched." + value_name + "=d.key");
        }
        sql_ = "SELECT k.id,"
                      "obr.uuid," +
                       what.get_sum() + " "
               "FROM (SELECT " + searched_what.get_sum() + ") searched(" + searched_as.get_sum() + ")," +
                    "keyset k "
               "JOIN dnskey d ON d.keysetid=k.id "
               "JOIN object_registry obr ON obr.id=k.id "
               "WHERE " + where.get_sum() + " "
               "ORDER BY 1";
    }
    const std::string& get_sql()const noexcept
    {
        return sql_;
    }
    static void push_back(const Database::Result::Row& row, std::vector<MatchedKeyset>& keysets_by_id)
    {
        static constexpr auto column_id = 0;
        static constexpr auto column_uuid = 1;
        static constexpr auto column_similarity_start = 2;
        MatchedKeyset match;
        match.id = static_cast<unsigned long long>(row[column_id]);
        const bool check_order = !keysets_by_id.empty();
        if (check_order)
        {
            const bool order_is_correct = keysets_by_id[keysets_by_id.size() - 1].id < match.id;
            if (!order_is_correct)
            {
                struct IncorrectOrder : std::runtime_error
                {
                    IncorrectOrder() : std::runtime_error("Incorrect order of matched keysets'") { }
                };
                throw IncorrectOrder();
            }
        }
        match.keyset_id = Util::make_strong<Registry::Keyset::KeysetId>(
            boost::uuids::string_generator()(static_cast<std::string>(row[column_uuid])));
        match.matched_items.insert(KeysetItem::public_key);
        match.similarity.reserve(row.size());
        for (decltype(row.size()) column_idx = column_similarity_start; column_idx < row.size(); ++column_idx)
        {
            match.similarity.push_back(static_cast<bool>(row[column_idx]) ? 1.0 : 0.0);
        }
        match.value_length = 0;
        keysets_by_id.push_back(match);
    }
private:
    std::string sql_;
};

std::unique_ptr<std::vector<MatchedKeyset>> combine(const std::vector<MatchedKeyset>& a, const std::vector<MatchedKeyset>& b)
{
    auto sum = std::make_unique<std::vector<MatchedKeyset>>();
    sum->reserve(a.size() + b.size());
    auto a_top = a.begin();
    auto b_top = b.begin();
    while ((a_top != a.end()) && (b_top != b.end()))
    {
        if (a_top->id < b_top->id)
        {
            sum->push_back(*a_top);
            ++a_top;
        }
        else if (b_top->id < a_top->id)
        {
            sum->push_back(*b_top);
            ++b_top;
        }
        else
        {
            auto c = *a_top;
            c.matched_items.insert(b_top->matched_items.begin(), b_top->matched_items.end());
            for (std::size_t idx = 0; idx < c.similarity.size(); ++idx)
            {
                if (c.similarity[idx] < b_top->similarity[idx])
                {
                    c.similarity[idx] = b_top->similarity[idx];
                }
            }
            sum->push_back(c);
            ++a_top;
            ++b_top;
        }
    }
    while (a_top != a.end())
    {
        sum->push_back(*a_top);
        ++a_top;
    }
    while (b_top != b.end())
    {
        sum->push_back(*b_top);
        ++b_top;
    }
    return sum;
}

template <KeysetItem ...items>
using KeysetColumns = std::integer_sequence<KeysetItem, items...>;

template <typename T>
using SearchableItems = Util::EnumSequence<KeysetColumns<KeysetItem::keyset_handle,
                                                         KeysetItem::public_key>,
                                           T>;

class RunQuery
{
public:
    RunQuery(const std::set<KeysetItem>& searched_items,
             const Database::QueryParams& params)
        : searched_items_{searched_items},
        params_{params}
    { }
    template <KeysetItem item>
    void operator()(Util::SqlInThreadSolver& running_sql)const
    {
        const bool to_search = 0 < searched_items_.count(item);
        if (to_search)
        {
            running_sql = Util::sql_exec_in_thread(SearchIn<item>{params_.size()}.get_sql(), params_);
        }
    }
private:
    const std::set<KeysetItem>& searched_items_;
    const Database::QueryParams& params_;
};

class GetResult
{
public:
    GetResult()
        : sum_matches_{std::make_unique<std::vector<MatchedKeyset>>()}
    { }
    template <KeysetItem column>
    void operator()(Util::SqlInThreadSolver& running_sql)
    {
        if (running_sql.is_joinable())
        {
            const Database::Result task_result = Util::SqlInThreadSolver::Result(std::move(running_sql)).get_result();
            if (0 < task_result.size())
            {
                auto matches = std::make_unique<std::vector<MatchedKeyset>>();
                matches->reserve(task_result.size());
                for (std::size_t row_idx = 0; row_idx < task_result.size(); ++row_idx)
                {
                    SearchIn<column>::push_back(task_result[row_idx], *matches);
                }
                if (sum_matches_->empty())
                {
                    sum_matches_ = std::move(matches);
                }
                else
                {
                    auto combined_matches = combine(*sum_matches_, *matches);
                    sum_matches_ = std::move(combined_matches);
                }
            }
        }
    }
    const std::vector<MatchedKeyset>& get_sum_matches()const noexcept
    {
        return *sum_matches_;
    }
    void release_sum_matches()noexcept
    {
        sum_matches_.release();
    }
private:
    std::unique_ptr<std::vector<MatchedKeyset>> sum_matches_;
};

}//namespace Fred::Registry::Keyset::{anonymous}

SearchKeysetReply::Data search_keyset(const SearchKeysetRequest& request)
{
    const Database::QueryParams params(request.query_values.begin(), request.query_values.end());
    Database::QueryParams params_ip_addresses;
    SearchableItems<Util::SqlInThreadSolver> running_sql{};
    visit(running_sql, RunQuery(request.searched_items, params));
    GetResult data_collector{};
    {
        const struct LogDuration : Util::MeasureDuration<LogDuration>
        {
            static void measure_completed(Time duration_ms)
            {
                LOGGER.debug("data collected in " + std::to_string(duration_ms) + "ms");
            }
        } log_duration{};
        visit(running_sql, data_collector);
    }
    const struct LogDuration : Util::MeasureDuration<LogDuration>
    {
        static void measure_completed(Time duration_ms)
        {
            LOGGER.debug("matches completed in " + std::to_string(duration_ms) + "ms");
        }
    } log_duration{};
    struct ResultWithSimilarity
    {
        explicit ResultWithSimilarity(const MatchedKeyset& src)
            : similarity(compute_similarity(src)),
              value_length{src.value_length}
        {
            if (0.0 < similarity)
            {
                result.keyset_id = src.keyset_id;
                result.matched_items = src.matched_items;
            }
        }
        SearchKeysetReply::Data::Result result;
        double similarity;
        std::size_t value_length;
        static double compute_similarity(const MatchedKeyset& src)
        {
            if (src.similarity.size() == 1)
            {
                return src.similarity[0];
            }
            double sum = 0;
            for (const auto similarity : src.similarity)
            {
                if (similarity <= 0.0)
                {
                    return 0.0;
                }
                sum += similarity * similarity;
            }
            return std::sqrt(sum / src.similarity.size());
        }
    };
    std::vector<ResultWithSimilarity> matches_by_similarity;
    matches_by_similarity.reserve(data_collector.get_sum_matches().size());
    for (const auto& matched_contact : data_collector.get_sum_matches())
    {
        const ResultWithSimilarity item(matched_contact);
        if (0.0 < item.similarity)
        {
            matches_by_similarity.push_back(item);
        }
    }
    data_collector.release_sum_matches();
    struct BySimilarityDescByLengthAsc
    {
        bool operator()(const ResultWithSimilarity& lhs, const ResultWithSimilarity& rhs)const
        {
            return (rhs.similarity < lhs.similarity) ||
                   ((rhs.similarity == lhs.similarity) && (lhs.value_length < rhs.value_length));
        }
    };
    std::sort(matches_by_similarity.begin(), matches_by_similarity.end(), BySimilarityDescByLengthAsc());
    SearchKeysetReply::Data result;
    result.searched_items = request.searched_items;
    result.result_count.lower_estimate = matches_by_similarity.size();
    result.result_count.upper_estimate = matches_by_similarity.size();
    const bool unlimited = request.limit == boost::none;
    if (unlimited || (matches_by_similarity.size() <= *request.limit))
    {
        result.most_similar_keysets.reserve(matches_by_similarity.size());
        for (const auto& item : matches_by_similarity)
        {
            result.most_similar_keysets.push_back(item.result);
        }
        return result;
    }
    result.most_similar_keysets.reserve(*request.limit);
    for (const auto& item : matches_by_similarity)
    {
        if (*request.limit <= result.most_similar_keysets.size())
        {
            return result;
        }
        result.most_similar_keysets.push_back(item.result);
    }
    return result;
}

}//namespace Fred::Registry::Keyset
}//namespace Fred::Registry
}//namespace Fred
